-- Public Domain software.
-- Euphoria 2.5 -- unlimited alphabet bilingual interpreter.
-- Any code page for the names.
-- English and Russian keywords, builtins, standard routines.
-- English  or Russian interpreter messages.
-- A Euphoria Interpreter written 100% in Euphoria,
-- reconstruction by Igor N. Kachan, 2005 May.
-- Report the problems to  kinz@peterlink.ru.

-- The Interpreter Back-End

-- This back-end is written in Euphoria. It uses the same front-end
-- as the official RDS Euphoria interpreter, and it executes the same IL
-- opcodes. Because it's written in Euphoria, this back-end is very 
-- simple, much simpler than the back-end used in the 
-- official RDS interpreter. 

-- Using the Euphoria to C Translator, or the Binder, you can convert 
-- this 100% Euphoria-coded interpreter into a .exe. The Translator 
-- will boost its speed considerably, though it will still be slower 
-- than the official RDS interpreter. The official interpreter has a 
-- carefully hand-coded back-end written in C.

include dll.e
include file.e
include global_r.e
include opnamesr.e

-- we handle these operations specially because they refer to routine ids
-- in the user program, not the interpreter itself. We can't just let 
-- Euphoria do the work.
constant M_CALL_BACK = 52,  
	 M_CRASH_ROUTINE = 66,
	 M_CRASH_MESSAGE = 37,
	 M_CRASH_FILE = 57
	 
constant C_MY_ROUTINE = 1,
	 C_USER_ROUTINE = 2,
	 C_NUM_ARGS = 3

object crash_msg
crash_msg = 0

sequence call_backs, call_back_code
symtab_index t_id, t_arglist, t_return_val, call_back_routine

sequence crash_list  -- list of routine id's to call if there's a fatal crash
crash_list = {}

integer crash_count
crash_count = 0

-- only need one set of temps for call-backs
t_id = tmp_alloc()
t_arglist = tmp_alloc()
t_return_val = tmp_alloc()

-- dummy call-back routine
call_back_routine = NewEntry("_call_back_", 0, 0, PROC, 0, 0, 0)
SymTab[call_back_routine] = SymTab[call_back_routine] & 
			    repeat(0, SIZEOF_ROUTINE_ENTRY - 
			    length(SymTab[call_back_routine]))

call_back_code = {CALL_FUNC,
		  t_id,
		  t_arglist,
		  t_return_val,
		  CALL_BACK_RETURN
		 }

SymTab[call_back_routine][S_CODE] = call_back_code

integer TraceOn
TraceOn = FALSE

integer pc, a, b, c, d, target, len, keep_running
integer lhs_seq_index -- index of lhs sequence
sequence lhs_subs -- first n-1 LHS subscripts before final subscript or slice
sequence val

sequence call_stack -- subroutine call stack

-- Stack Frames

-- 1. recursive call (routine is already active):

-- arg1 arg2 arg3 ... private1 private2 ... temp1 temp2 ... 
--                              ... previous_stack_size caller_pc current_sub

-- 2. non-recursive call:

-- previous_stack_size caller_pc current_sub

integer err_file
sequence err_file_name
err_file_name = "ex.err" 

procedure open_err_file()
-- open ex.err  

    err_file = open(err_file_name, "w")
    if err_file = -1 then
	puts(2, messages[124][LANG] & err_file_name & '\n')--"Can't open "
	abort(1)
    end if
end procedure

procedure both_puts(object s)
-- print to both screen and error file  
    if atom(crash_msg) then
	puts(2, s)
    end if
    puts(err_file, s)
end procedure

procedure both_printf(sequence format, sequence items)
-- print to both screen and error file  
    if atom(crash_msg) then
	printf(2, format, items)
    end if
    printf(err_file, format, items)
end procedure

function find_line(symtab_index sub, integer pc)
-- return the file name and line that matches pc in sub
    sequence linetab
    integer line, gline
    
    linetab = SymTab[sub][S_LINETAB]
    line = 1
    for i = 1 to length(linetab) do
	if linetab[i] >= pc or linetab[i] = -2 then
	    line = i-1
	    while line > 1 and linetab[line] = -1 do
		line -= 1
	    end while
	    exit
	end if
    end for
    gline = SymTab[sub][S_FIRSTLINE] + line - 1
    return {file_name[slist[gline][LOCAL_FILE_NO]], slist[gline][LINE]}
end function

procedure show_var(symtab_index x)
-- display a variable name and value
    
    puts(err_file, "    " & SymTab[x][S_NAME] & " = ")
    if equal(val[x], NOVALUE) then
	puts(err_file, messages[125][LANG])--"<no value>"
    else
	pretty_print(err_file, val[x], 
	{1, 2, length(SymTab[x][S_NAME]) + 7, 78, "%d", "%.10g", 32, 127, 500})
    end if
    puts(err_file, '\n')
end procedure

procedure trace_back(sequence msg)--kinz
-- display the call stack and variables after a crash
    symtab_index sub, v
    integer levels, prev_file_no, sp

    levels = 1
    if atom(slist[$]) then
	slist = s_expand(slist)
    end if
    while TRUE do
	sub = call_stack[$]
	
	if levels = 1 then
	    puts(2, '\n')
	
	elsif sub != call_back_routine then
	    both_puts(messages[126][LANG])--"... called from "
	    -- pc points to statement after the subroutine call
	end if
	
	if sub = call_back_routine then
	    if crash_count > 0 then
	both_puts(messages[127][LANG])--"^^^ called to handle run-time crash\n"
		exit
	    else
		both_puts(messages[128][LANG])--"^^^ call-back from "
		if EWINDOWS then
		    both_puts(messages[129][LANG])--"Windows\n"
		else    
		    both_puts(messages[130][LANG])--"external program\n"
		end if
	    end if
	
	else
	    both_printf("%s:%d ", find_line(sub, pc)) 

	    if not equal(SymTab[sub][S_NAME], "_toplevel_") then
		if SymTab[sub][S_TOKEN] = PROC then
		    both_puts(messages[131][LANG])--"in procedure "
		elsif SymTab[sub][S_TOKEN] = FUNC then
		    both_puts(messages[132][LANG])--"in function "
		elsif SymTab[sub][S_TOKEN] = TYPE then
		    both_puts(messages[133][LANG])--"in type "
		end if
	
		both_printf("%s()", {SymTab[sub][S_NAME]})
	    end if
	    
	    both_puts("\n")
	    
	    if levels = 1 then
		if sequence(crash_msg) then
		    clear_screen()
		    puts(2, crash_msg)
		end if
		both_puts(msg & '\n')
	    end if

	    if length(call_stack) < 2 then
		both_puts('\n')
		exit
	    end if
	    
	    -- display parameters and private vars
	    v = SymTab[sub][S_NEXT]
	    sp = call_stack[$-2]+1
	    while v != 0 and 
		(SymTab[v][S_SCOPE] = SC_PRIVATE or 
		SymTab[v][S_SCOPE] = SC_LOOP_VAR) do
		show_var(v)
		if SymTab[sub][S_ACTIVE] > 1 then
		    -- recursive call, restore private var
		    val[v] = call_stack[sp]
		    sp += 1
		end if
		v = SymTab[v][S_NEXT]
	    end while
	    SymTab[sub][S_ACTIVE] -= 1
	end if
	
	puts(err_file, '\n')
	
	-- stacked pc points to next statement after the call (so subtract 1)
	pc = call_stack[$-1] - 1
	call_stack = call_stack[1..call_stack[$-2]]
	levels += 1
    end while
    
    puts(2, messages[134][LANG] & err_file_name & '\n')-- "--> see "
    
    puts(err_file, messages[135][LANG] )--"\n\nGlobal & Local Variables\n"
    prev_file_no = -1
    v = SymTab[TopLevelSub][S_NEXT]
    while v do
	if SymTab[v][S_TOKEN] = VARIABLE and 
	   SymTab[v][S_MODE] = M_NORMAL and
	   find(SymTab[v][S_SCOPE], {SC_LOCAL, SC_GLOBAL, SC_GLOOP_VAR}) then
	    if SymTab[v][S_FILE_NO] != prev_file_no then
		prev_file_no = SymTab[v][S_FILE_NO]
		puts(err_file, "\n " & file_name[prev_file_no] & ":\n")
	    end if
	    show_var(v)
	end if
	v = SymTab[v][S_NEXT]
    end while
    puts(err_file, '\n')
    close(err_file)
end procedure

integer forward_general_callback, forward_machine_callback

procedure call_crash_routines()
-- call all the routines in the crash list  
    object quit
    
    if crash_count > 0 then
	return
    end if
    
    crash_count += 1
    
    -- call them in reverse order
    err_file_name = "ex_crash.err"
    
    for i = length(crash_list) to 1 by -1 do
	-- do callback to get addr
	quit = call_func(forward_general_callback, 
			 {{0, crash_list[i], 1}, {0}})
	if not equal(quit, 0) then
	    return -- don't call the others
	end if
    end for
end procedure

procedure quit_after_error()
-- final termination    
    if EWINDOWS then
	puts(2,messages[136][LANG])--"\nPress Enter...\n"
	if getc(0) then
	end if
    end if
    abort(1)
end procedure

procedure RTFatalType(integer x) --kinz
-- handle a fatal run-time type-check error 
    sequence msg, v
    sequence vname
    
    open_err_file()
    a = Code[x]
    vname = SymTab[a][S_NAME]
    msg = sprintf(messages[137][LANG], {vname}) --"type_check error\n%s is "
    v = sprint(val[a])
    if length(v) > 70 - length(vname) then
	v = v[1..70 - length(vname)]
	while length(v) and not find(v[$], ",}")  do
	    v = v[1..$-1]
	end while
	v = v & " ..."
    end if
    trace_back(msg & v)
    call_crash_routines()
    quit_after_error()
end procedure

procedure RTFatal(object msg) -- kinz
-- handle a fatal run-time error    
    open_err_file()
      if atom(msg) then
	  trace_back(messages[msg][LANG])
      else
	  trace_back(msg)
      end if
    call_crash_routines()
    quit_after_error()
end procedure

procedure RTInternal(sequence msg)
-- Internal errors in back-end
    puts(2, '\n' & msg & '\n')
    ? 1/0
end procedure

file trace_file
trace_file = -1

integer trace_line
trace_line = 0

procedure one_trace_line(sequence line)
-- write one fixed-width 79-char line to ctrace.out
    if ELINUX then
	printf(trace_file, "%-78.78s\n", {line})
    else
	printf(trace_file, "%-77.77s\r\n", {line})
    end if
end procedure

procedure opSTARTLINE() --kinz
-- Start of a line. Use for diagnostics.
    sequence line
    integer w
    
    if TraceOn then
	if trace_file = -1 then
	    trace_file = open("ctrace.out", "wb")
	    if trace_file = -1 then
		RTFatal(138)--"Couldn't open ctrace.out"
	    end if
	end if

	a = Code[pc+1]
	
	if atom(slist[$]) then
	    slist = s_expand(slist)
	end if
	line = fetch_line(slist[a][SRC])
	line = sprintf("%s:%d\t%s",
		       {name_ext(file_name[slist[a][LOCAL_FILE_NO]]),
			slist[a][LINE],
			line})
	trace_line += 1
	if trace_line >= 500 then
	    -- wrap around to start of file
	    trace_line = 0
	    one_trace_line("")
	    one_trace_line("               ")
	    flush(trace_file)
	    if seek(trace_file, 0) then
	    end if
	end if
	
	one_trace_line(line)
	one_trace_line("")
	one_trace_line(messages[139][LANG])--"=== THE END ==="
	one_trace_line("")
	one_trace_line("")
	one_trace_line("")
	flush(trace_file)
	w = where(trace_file)
	if seek(trace_file, w-79*5) then -- back up 5 (fixed-width) lines
	end if
    end if
    pc += 2
end procedure
	
procedure opPROC()  
-- Normal subroutine call
    integer n, arg, sub, top
    
    -- make a procedure or function/type call
    sub = Code[pc+1] -- subroutine
    arg = SymTab[sub][S_NEXT] 
    n = SymTab[sub][S_NUM_ARGS]
    SymTab[sub][S_ACTIVE] += 1
    top = length(call_stack)
    
    if SymTab[sub][S_ACTIVE] > 1 then
	-- recursive: push the params, privates and temps
	--- save and set args
	for i = 1 to n do
	    call_stack = append(call_stack, val[arg])   
	    val[arg] = val[Code[pc+1+i]]
	    arg = SymTab[arg][S_NEXT]
	end for
	
	-- save privates
	while arg != 0 and SymTab[arg][S_SCOPE] <= SC_PRIVATE do
	    call_stack = append(call_stack, val[arg])   
	    val[arg] = NOVALUE  
	    arg = SymTab[arg][S_NEXT]
	end while
	
	-- save temps
	arg = SymTab[sub][S_TEMPS]
	while arg != 0 do
	    call_stack = append(call_stack, val[arg])   
	    val[arg] = NOVALUE
	    arg = SymTab[arg][S_NEXT]
	end while
    else    
	-- non recursive, no need to push privates or temps
	-- copy args
	for i = 1 to n do
	    val[arg] = val[Code[pc+1+i]]
	    arg = SymTab[arg][S_NEXT]
	end for
    end if
    
    call_stack = append(call_stack, top)    

    pc = pc + 2 + n
    if SymTab[sub][S_TOKEN] != PROC then
	pc += 1
    end if
    
    call_stack = append(call_stack, pc) 
    call_stack = append(call_stack, sub)
    
    Code = SymTab[sub][S_CODE]
    pc = 1 
end procedure

integer result
result = 0
object result_val

procedure opRETURNP()   
-- return from procedure (or function)
    integer sub, arg, top, base
    
    sub = Code[pc+1]
    arg = SymTab[sub][S_NEXT]
    SymTab[sub][S_ACTIVE] -= 1
    top = call_stack[$-2]
    
    if SymTab[sub][S_ACTIVE] > 0 then
	-- restore the parameters and privates
	base = top+1
	while arg and SymTab[arg][S_SCOPE] <= SC_PRIVATE do
	    val[arg] = call_stack[base]
	    base += 1
	    arg = SymTab[arg][S_NEXT]
	end while
	
	-- restore the temps
	arg = SymTab[sub][S_TEMPS]
	while arg != 0 do
	    val[arg] = call_stack[base]
	    base += 1
	    arg = SymTab[arg][S_NEXT]
	end while
    else
	-- don't restore the privates or temps
	-- but set privates to NOVALUE
	while arg and SymTab[arg][S_SCOPE] <= SC_PRIVATE do
	    val[arg] = NOVALUE
	    arg = SymTab[arg][S_NEXT]
	end while
    
    end if
    
    pc = call_stack[$-1]
    call_stack = call_stack[1..top]
    
    Code = SymTab[call_stack[$]][S_CODE]
    
    if result then
	val[Code[result]] = result_val
	result = 0
    end if
end procedure

procedure opRETURNF()  
-- return from function
    result_val = val[Code[pc+2]]
    result = call_stack[$-1] - 1
    opRETURNP()
end procedure

procedure opCALL_BACK_RETURN()
-- force return from do_exec()
    keep_running = FALSE  
end procedure
		
procedure opBADRETURNF()--kinz
-- shouldn't reach here
  RTFatal(140)--"attempt to exit a function without returning a value"
end procedure

procedure opRETURNT()   
-- return from top-level "procedure"
    pc += 1
    if pc > length(Code) then
	keep_running = FALSE  -- we've reached the end of the code
    end if
end procedure
	
procedure opRHS_SUBS() 
-- subscript a sequence to get the value of the element
-- RHS_SUBS_CHECK, RHS_SUBS, RHS_SUBS_I
    object sub, x
    
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    x = val[a]
    sub = val[b]
    if atom(x) then
       RTFatal(141)--"attempt to subscript an atom\n(reading from it)"
    end if
    if sequence(sub) then
   RTFatal(142)--"subscript must be an atom\n(reading an element of a sequence)"
    end if
    if sub < 1 or sub > length(x) then
RTFatal(sprintf(messages[143][LANG], 
--"subscript value %d is out of bounds, reading from a sequence of length %d"   
	{sub, length(x)}))
    end if
    val[target] = x[sub]
    pc += 4
end procedure
	
procedure opIF()
    a = Code[pc+1]
    if val[a] = 0 then
	pc = Code[pc+2]
    else
	pc += 3
    end if
end procedure
	    
procedure opINTEGER_CHECK()
    a = Code[pc+1]
    if not integer(val[a]) then
	RTFatalType(pc+1)
    end if
    pc += 2
end procedure
	    
procedure opATOM_CHECK()
    a = Code[pc+1]
    if not atom(val[a]) then
	RTFatalType(pc+1)
    end if
    pc += 2
end procedure
	      
procedure opSEQUENCE_CHECK()
    a = Code[pc+1]
    if not sequence(val[a]) then
	RTFatalType(pc+1)
    end if
    pc += 2
end procedure

procedure opASSIGN()  
-- ASSIGN, ASSIGN_I 
    a = Code[pc+1]
    target = Code[pc+2]
    val[target] = val[a]
    pc += 3
end procedure
		
procedure opELSE()  
-- ELSE, EXIT, ENDWHILE
    pc = Code[pc+1]
end procedure
	    
procedure opRIGHT_BRACE_N()  
-- form a sequence of any length 
    sequence x
    
    len = Code[pc+1]
    x = {}
    for i = pc+len+1 to pc+2 by -1 do
	-- last one comes first
	x = append(x, val[Code[i]])
    end for
    target = Code[pc+len+2]
    val[target] = x
    pc += 3 + len
end procedure

procedure opRIGHT_BRACE_2()   
-- form a sequence of length 2 (slightly faster than above)
    target = Code[pc+3]
    -- the second one comes first
    val[target] = {val[Code[pc+2]], val[Code[pc+1]]}
    pc += 4
end procedure

procedure opPLUS1() 
--PLUS1, PLUS1_I
    a = Code[pc+1]
    -- [2] is not used
    target = Code[pc+3] 
    val[target] = val[a] + 1
    pc += 4
end procedure
	
procedure opGLOBAL_INIT_CHECK()  
-- GLOBAL_INIT_CHECK, PRIVATE_INIT_CHECK 
    a = Code[pc+1]
    if equal(val[a], NOVALUE) then
	RTFatal(144) -- "not initialized" -- ?? 
    end if
    pc += 2
end procedure
	    
procedure opWHILE()     
-- sometimes emit.c optimizes this away
    a = Code[pc+1]
    if val[a] = 0 then
	pc = Code[pc+2]
    else
	pc += 3
    end if
end procedure

function var_subs(object x, sequence subs)
-- subscript x with the list of subscripts in subs
    if atom(x) then
       RTFatal(141)--"attempt to subscript an atom\n(reading from it)"
    end if
    for i = 1 to length(subs) do
	if subs[i] > length(x) or subs[i] < 1 then
	    RTFatal(sprintf(messages[143][LANG],
--"subscript value %d is out of bounds, reading from a sequence of length %d"
		{subs[i], length(x)}))
	end if
	x = x[subs[i]]
    end for
    return x
end function

function var_slice(object x, sequence subs, atom lower, atom upper)
-- slice x after subscripting a variable number of times
    if atom(x) then
       RTFatal(141)--"attempt to subscript an atom\n(reading from it)"
    end if
    for i = 1 to length(subs) do
	if subs[i] > length(x) or subs[i] < 1 then
       RTFatal(sprintf(messages[143][LANG],
--"subscript value %d is out of bounds, reading from a sequence of length %d"
		{subs[i], length(x)}))
	end if
	x = x[subs[i]] 
    end for
    return x[lower..upper]
end function

procedure opLENGTH()
-- operand should be a sequence 
    a = Code[pc+1]
    target = Code[pc+2]
    if atom(val[a]) then
	RTFatal(145)--"length of an atom is not defined"
    end if
    val[target] = length(val[a])
    pc += 3
end procedure

-- Note: Multiple LHS subscripts, and $ within those subscripts,
-- is handled much more efficiently in the hand-coded C interpreter,
-- and in code translated to C, where C pointers can be used effectively.

procedure opPLENGTH()
-- Needed for some LHS uses of $. Operand should be a val index of a sequence,
-- with subscripts. 
    a = Code[pc+1]
    target = Code[pc+2]
    lhs_seq_index = val[a][1]
    lhs_subs = val[a][2..$]
    val[target] = length(var_subs(val[lhs_seq_index], lhs_subs))
    pc += 3
end procedure

procedure opLHS_SUBS() 
-- LHS = "Left Hand Side" of assignment
-- Handle one LHS subscript, when there are multiple LHS subscripts.
    
    a = Code[pc+1] -- base var sequence, or a temp that contains
		   -- {base index, subs1, subs2... so far}
    b = Code[pc+2] -- subscript
    target = Code[pc+3] -- temp for storing result
    
    -- a is a "pointer" to the result of previous subscripting
    val[target] = append(val[a], val[b])
    pc += 5
end procedure

procedure opLHS_SUBS1() 
-- Handle first LHS subscript, when there are multiple LHS subscripts.
    a = Code[pc+1] -- base var sequence, or a temp that contains
		   -- {base index, subs1, subs2... so far}
    b = Code[pc+2] -- subscript
    target = Code[pc+3] -- temp for storing result
    
    -- a is the base var
    val[target] = {a, val[b]}
    pc += 5
end procedure

procedure opLHS_SUBS1_COPY() 
-- Handle first LHS subscript, when there are multiple LHS subscripts.
-- In tricky situations (in the C-coded back-end) a copy of the sequence 
-- is made into a temp. 
    
    a = Code[pc+1] -- base var sequence
		   
    b = Code[pc+2] -- subscript
    
    target = Code[pc+3] -- temp for storing result
    
    c = Code[pc+4] -- temp to hold base sequence while it's manipulated
    
    val[c] = val[a]
    
    -- a is the base var
    val[target] = {c, val[b]}
    
    pc += 5
end procedure

procedure lhs_check_subs(object seq, object subs)
-- see if seq[subs] = ... is legal  
    if atom(seq) then
	RTFatal(146)--"attempt to subscript an atom\n(assigning to it)"
    end if
    if sequence(subs) then
	RTFatal(sprintf(messages[147][LANG],
--"subscript must be an atom\n(assigning to a sequence of length %d)"
	length(seq)))
    end if
    if subs < 1 or subs > length(seq) then
	RTFatal(sprintf(messages[148][LANG],
--"subscript value %d is out of bounds, assigning to a sequence of length %d"
	{subs, length(seq)}))
    end if
end procedure

procedure lhs_check_slice(object seq, object lower, object upper, object rhs)
-- check for a valid assignment to a slice
    atom len
    
    if sequence(lower) then
	RTFatal(149)--"slice lower index is not an atom"
    end if
    if lower < 1 then
	RTFatal(150)--"slice lower index is less than 1"
    end if
    
    if sequence(upper) then
	RTFatal(151)--"slice upper index is not an atom"
    end if
    if upper < 0 then
	RTFatal(152)--"slice upper index is less than 0"
    end if
    
    len = upper - lower + 1
    
    if len < 0 then
	RTFatal(153)--"slice length is less than 0"
    end if
    
    if sequence(rhs) and length(rhs) != len then
	RTFatal(154)--"lengths do not match on assignment to slice"
    end if

    if atom(seq) then
	RTFatal(155)--"attempt to slice an atom"
    end if
	
    if lower > length(seq) + 1 or (len > 0 and lower > length(seq)) then
	RTFatal(156)--"slice starts past end of sequence"
    end if
	
    if upper > length(seq) then
	RTFatal(157)--"slice ends past end of sequence"
    end if
end procedure

function assign_subs(sequence x, sequence subs, object rhs_val)
-- assign a value to a subscripted sequence (any number of subscripts >= 1)
    lhs_check_subs(x, subs[1])
    if length(subs) = 1 then
	x[subs[1]] = rhs_val
    else
	x[subs[1]] = assign_subs(x[subs[1]], subs[2..$], rhs_val)
    end if
    return x
end function            

function assign_slice(sequence x, sequence subs, atom lower, atom upper, object rhs_val)
-- assign a value to a subscripted/sliced sequence 
-- (any number of subscripts >= 1, then one slice)
    -- should check slice too
    lhs_check_subs(x, subs[1])
    if length(subs) = 1 then
	lhs_check_slice(x[subs[1]],lower,upper,rhs_val)
	x[subs[1]][lower..upper] = rhs_val
    else
	x[subs[1]] = assign_slice(x[subs[1]], subs[2..$], lower, upper, rhs_val)
    end if
    return x
end function            

procedure opASSIGN_SUBS() -- also ASSIGN_SUBS_CHECK, ASSIGN_SUBS_I
-- LHS single subscript and assignment
    object x, subs
    
    a = Code[pc+1]  -- the sequence
    b = Code[pc+2]  -- the subscript
    if sequence(val[b]) then
      RTFatal(158)
      --"subscript must be an atom\n(assigning to subscript of a sequence)"
    end if
    c = Code[pc+3]  -- the RHS value
    x = val[a] -- avoid lingering ref count on val[a]
    lhs_check_subs(x, val[b])
    x = val[c]
    subs = val[b]
    val[a][subs] = x  -- single LHS subscript
    pc += 4
end procedure

procedure opPASSIGN_SUBS()
-- final LHS subscript and assignment after a series of subscripts
    a = Code[pc+1]
    b = Code[pc+2]  -- subscript
    if sequence(val[b]) then
	RTFatal(159)
	--"subscript must be an atom\n(assigning to subscript of a sequence)"
    end if
    c = Code[pc+3]  -- RHS value
    
    -- multiple LHS subscript case
    lhs_seq_index = val[a][1]
    lhs_subs = val[a][2..$]    
    val[lhs_seq_index] = assign_subs(val[lhs_seq_index], 
					 lhs_subs & val[b], 
					 val[c])
    pc += 4
end procedure

procedure opASSIGN_OP_SUBS()  
-- var[subs] op= expr
    object x
    
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    -- var with one subscript
    lhs_subs = {}
    x = val[a]
    val[target] = var_subs(x, lhs_subs & val[b])
    pc += 4
end procedure

procedure opPASSIGN_OP_SUBS()  
-- var[subs] ... [subs] op= expr
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    -- temp with multiple subscripts
    lhs_seq_index = val[a][1]
    lhs_subs = val[a][2..$]
    Code[pc+9] = Code[pc+1] -- patch upcoming op
    val[target] = var_subs(val[lhs_seq_index], lhs_subs & val[b])
    pc += 4
end procedure

procedure opASSIGN_OP_SLICE()  
-- var[i..j] op= expr
    object x

    a = Code[pc+1]
    x = val[a]
    b = Code[pc+2]
    if val[b] > length(x) or val[b] < 1 then
	RTFatal(sprintf(messages[143][LANG],
--"subscript value %d is out of bounds, reading from a sequence of length %d"
		{val[b], length(x)}))
    end if
    c = Code[pc+3]
    target = Code[pc+4]
    val[target] = var_slice(x, {}, val[b], val[c])
    pc += 5
end procedure

-- procedure opPASSIGN_OP_SLICE() -- buggy variant reported by Vincent.
-- -- var[subs] ... [i..j] op= expr
--  object x
-- 
--  a = Code[pc+1]
--  x = val[a]
--  b = Code[pc+2]
--  if val[b] > length(x) or val[b] < 1 then
--      RTFatal(sprintf(messages[143][LANG],
-- --"subscript value %d is out of bounds, reading from a sequence of length %d"
--              {val[b], length(x)}))
--  end if
--  c = Code[pc+3]
--  target = Code[pc+4]
--  lhs_seq_index = val[a][1]
--  lhs_subs = val[a][2..length(val[a])]
--  Code[pc+10] = Code[pc+1]
--  val[target] = var_slice(val[lhs_seq_index], lhs_subs, val[b], val[c])
--  pc += 5
-- end procedure

procedure opPASSIGN_OP_SLICE() -- corrected by Rob Craig in EUForum message.
-- var[subs] ... [i..j] op= expr
    object x

    a = Code[pc+1]
    x = val[a]
    b = Code[pc+2]
    c = Code[pc+3]
    target = Code[pc+4]
    lhs_seq_index = x[1]
    lhs_subs = x[2..$]
    Code[pc+10] = Code[pc+1]
    val[target] = var_slice(val[lhs_seq_index], lhs_subs, val[b], val[c])
    pc += 5
end procedure
	    
procedure opASSIGN_SLICE()   
-- var[i..j] = expr
    object x
    
    a = Code[pc+1]  -- sequence
    b = Code[pc+2]  -- 1st index
    c = Code[pc+3]  -- 2nd index
    d = Code[pc+4]  -- rhs value to assign
    
    x = val[a] -- avoid lingering ref count on val[a]
    lhs_check_slice(x,val[b],val[c],val[d])
    x = val[d]
    val[a][val[b]..val[c]] = x
    pc += 5
end procedure

procedure opPASSIGN_SLICE()   
-- var[x] ... [i..j] = expr
    a = Code[pc+1]  -- sequence
    b = Code[pc+2]  -- 1st index
    c = Code[pc+3]  -- 2nd index
    d = Code[pc+4]  -- rhs value to assign
    
    lhs_seq_index = val[a][1]
    lhs_subs = val[a][2..$]
    val[lhs_seq_index] = assign_slice(val[lhs_seq_index],
				      lhs_subs,
				      val[b], val[c], val[d])
    pc += 5
end procedure

procedure opRHS_SLICE() 
-- rhs slice of a sequence a[i..j] 
    object x
    
    a = Code[pc+1]  -- sequence
    b = Code[pc+2]  -- 1st index
    c = Code[pc+3]  -- 2nd index
    target = Code[pc+4]  
    x = val[a]
    val[target] = x[val[b]..val[c]]   -- need checks here!
    pc += 5
end procedure

procedure opTYPE_CHECK() 
-- type check for a user-defined type
-- this always follows a type-call
    if val[Code[pc-1]] = 0 then
	RTFatalType(pc-2)
    end if
    pc += 1
end procedure
	    
procedure opIS_AN_INTEGER()
    a = Code[pc+1]
    target = Code[pc+2]
    val[target] = integer(val[a])
    pc += 3
end procedure

procedure opIS_AN_ATOM()
    a = Code[pc+1]
    target = Code[pc+2]
    val[target] = atom(val[a])
    pc += 3
end procedure
		
procedure opIS_A_SEQUENCE() 
    a = Code[pc+1]
    target = Code[pc+2]
    val[target] = sequence(val[a])
    pc += 3
end procedure
	    
procedure opIS_AN_OBJECT()
    target = Code[pc+2]
    val[target] = 1
    pc += 3
end procedure
		
	-- ---------- start of unary ops ----------------- 

procedure opSQRT() 
    a = Code[pc+1]
    target = Code[pc+2]
    val[target] = sqrt(val[a])
    pc += 3
end procedure

procedure opSIN()
    a = Code[pc+1]
    target = Code[pc+2]
    val[target] = sin(val[a])
    pc += 3
end procedure

procedure opCOS()
    a = Code[pc+1]
    target = Code[pc+2]
    val[target] = cos(val[a])
    pc += 3
end procedure

procedure opTAN()
    a = Code[pc+1]
    target = Code[pc+2]
    val[target] = tan(val[a])
    pc += 3
end procedure

procedure opARCTAN()
    a = Code[pc+1]
    target = Code[pc+2]
    val[target] = arctan(val[a])
    pc += 3
end procedure

procedure opLOG()
    a = Code[pc+1]
    target = Code[pc+2]
    val[target] = log(val[a])
    pc += 3
end procedure

procedure opNOT_BITS()
    a = Code[pc+1]
    target = Code[pc+2]
    val[target] = not_bits(val[a])
    pc += 3
end procedure

procedure opFLOOR()
    a = Code[pc+1]
    target = Code[pc+2]
    val[target] = floor(val[a])
    pc += 3
end procedure

procedure opNOT_IFW()
    a = Code[pc+1]
    if val[a] = 0 then
	pc += 3
    else
	pc = Code[pc+2]
    end if
end procedure
	    
procedure opNOT()
    a = Code[pc+1]
    target = Code[pc+2]
    val[target] = not val[a]
    pc += 3
end procedure
	    
procedure opUMINUS()
    a = Code[pc+1]
    target = Code[pc+2]
    val[target] = -val[a]
    pc += 3
end procedure
	    
procedure opRAND()
    a = Code[pc+1]
    target = Code[pc+2]
    val[target] = rand(val[a])
    pc += 3
end procedure
	    
procedure opDIV2()  
-- like unary op, but pc+=4
    a = Code[pc+1]
    -- Code[pc+2] not used
    target = Code[pc+3]
    val[target] = val[a] / 2
    pc += 4 
end procedure
	    
procedure opFLOOR_DIV2()
    a = Code[pc+1]
    -- Code[pc+2] not used
    target = Code[pc+3]
    val[target] = floor(val[a] / 2)
    pc += 4 
end procedure
		
	----------- start of binary ops ----------
	    
procedure opGREATER_IFW()
    a = Code[pc+1]
    b = Code[pc+2]
    if val[a] > val[b] then
	pc += 4
    else
	pc = Code[pc+3]
    end if
end procedure
	
procedure opNOTEQ_IFW()
    a = Code[pc+1]
    b = Code[pc+2]
    if val[a] != val[b] then
	pc += 4
    else
	pc = Code[pc+3]
    end if
end procedure
	
procedure opLESSEQ_IFW()
    a = Code[pc+1]
    b = Code[pc+2]
    if val[a] <= val[b] then
	pc += 4
    else
	pc = Code[pc+3]
    end if
end procedure
	
procedure opGREATEREQ_IFW()
    a = Code[pc+1]
    b = Code[pc+2]
    if val[a] >= val[b] then
	pc += 4
    else
	pc = Code[pc+3]
    end if
end procedure
	
procedure opEQUALS_IFW()
    a = Code[pc+1]
    b = Code[pc+2]
    if val[a] = val[b] then
	pc += 4
    else
	pc = Code[pc+3]
    end if
end procedure
	
procedure opLESS_IFW()
    a = Code[pc+1]
    b = Code[pc+2]
    if val[a] < val[b] then
	pc += 4
    else
	pc = Code[pc+3]
    end if
end procedure
	    
	-- other binary ops
	
procedure opMULTIPLY()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = val[a] * val[b]
    pc += 4
end procedure
	    
procedure opPLUS() 
-- PLUS, PLUS_I
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = val[a] + val[b]
    pc += 4
end procedure
	
procedure opMINUS() 
-- MINUS, MINUS_I
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = val[a] - val[b]
    pc += 4
end procedure
	    
procedure opOR()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = val[a] or val[b]
    pc += 4
end procedure
	
procedure opXOR()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = val[a] xor val[b]
    pc += 4
end procedure
	
procedure opAND()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = val[a] and val[b]
    pc += 4
end procedure
	    
procedure opDIVIDE()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    if equal(val[b], 0) then
	RTFatal(159)--"attempt to divide by 0"
    end if
    val[target] = val[a] / val[b]
    pc += 4
end procedure

procedure opREMAINDER()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    if equal(val[b], 0) then
       RTFatal(160)--"Can't get remainder of a number divided by 0"
    end if
    val[target] = remainder(val[a], val[b])
    pc += 4
end procedure
	    
procedure opFLOOR_DIV()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    if equal(val[b], 0) then
	RTFatal(159)--"attempt to divide by 0"
    end if
    val[target] = floor(val[a] / val[b])
    pc += 4
end procedure
	    
procedure opAND_BITS()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = and_bits(val[a], val[b])
    pc += 4
end procedure
	
procedure opOR_BITS()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = or_bits(val[a], val[b])
    pc += 4
end procedure
	
procedure opXOR_BITS()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = xor_bits(val[a], val[b])
    pc += 4
end procedure
	    
procedure opPOWER()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = power(val[a], val[b])
    pc += 4
end procedure
	    
procedure opLESS()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = val[a] < val[b]
    pc += 4
end procedure
	
procedure opGREATER()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = val[a] > val[b]
    pc += 4
end procedure
	
procedure opEQUALS()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = val[a] = val[b]
    pc += 4
end procedure
	
procedure opNOTEQ()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = val[a] != val[b]
    pc += 4
end procedure
	
procedure opLESSEQ()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = val[a] <= val[b]
    pc += 4
end procedure
	
procedure opGREATEREQ()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = val[a] >= val[b]
    pc += 4
end procedure
	    
-- short-circuit ops 

procedure opSC1_AND()
    a = Code[pc+1]
    b = Code[pc+2]
    if atom(val[a]) then
	if val[a] = 0 then
	    val[b] = 0
	    pc = Code[pc+3]
	    return
	end if
    else
	RTFatal(161)--"true/false condition must be an ATOM"
    end if
    pc += 4
end procedure

procedure opSC1_AND_IF() 
-- no need to store 0   
    a = Code[pc+1]
    b = Code[pc+2]
    if atom(val[a]) then
	if val[a] = 0 then
	    pc = Code[pc+3]
	    return
	end if
    else
	RTFatal(161)--"true/false condition must be an ATOM"
    end if
    pc += 4
end procedure

procedure opSC1_OR()
    a = Code[pc+1]
    b = Code[pc+2]
    if atom(val[a]) then
	if val[a] != 0 then
	    val[b] = 1
	    pc = Code[pc+3]
	    return
	end if
    else
	RTFatal(161)--"true/false condition must be an ATOM"
    end if
    pc += 4
end procedure
		
procedure opSC1_OR_IF()
-- no need to store 1
    a = Code[pc+1]
    b = Code[pc+2]
    if atom(val[a]) then
	if val[a] != 0 then
	    val[b] = 1
	    pc = Code[pc+3]
	    return
	end if
    else
	RTFatal(161)--"true/false condition must be an ATOM"
    end if
    pc += 4
end procedure
		
procedure opSC2_OR() 
-- SC2_OR,  SC2_AND 
-- short-circuit op
    a = Code[pc+1]
    b = Code[pc+2]
    if atom(val[a]) then
	val[b] = val[a] 
    else
	RTFatal(161)--"true/false condition must be an ATOM"
    end if
    pc += 3
end procedure

-- for loops 
	    
procedure opFOR()  
-- FOR, FOR_I 
-- enter into a for loop    
    integer increment, limit, initial, loopvar, jump
    
    increment = Code[pc+1]
    limit = Code[pc+2]
    initial = Code[pc+3]
    -- ignore current_sub = Code[pc+4] - we don't patch the ENDFOR
    -- so recursion is not a problem
    loopvar = Code[pc+5]
    jump = Code[pc+6]
    
    if sequence(val[initial]) then
	RTFatal(162)--"for-loop variable is not an atom"
    end if
    if sequence(val[limit]) then
	RTFatal(163)--"for-loop limit is not an atom"
    end if
    if sequence(val[increment]) then
	RTFatal(164)--"for-loop increment is not an atom"
    end if
    
    pc += 7 -- to enter into the loop
    
    if val[increment] >= 0 then
	-- going up
	if val[initial] > val[limit] then
	    pc = jump -- quit immediately, 0 iterations
	end if
    else
	-- going down
	if val[initial] < val[limit] then
	    pc = jump -- quit immediately, 0 iterations
	end if
    end if

    val[loopvar] = val[initial] -- initialize loop var
    
end procedure

procedure opENDFOR_GENERAL() 
-- ENDFOR_INT_UP, ENDFOR_UP, ENDFOR_INT_DOWN1,
-- ENDFOR_INT_DOWN, ENDFOR_DOWN, ENDFOR_GENERAL
-- end of for loop: drop out of the loop, or go back to the top
    integer loopvar
    atom increment, limit, next
    
    limit = val[Code[pc+2]]
    increment = val[Code[pc+4]]
    loopvar = Code[pc+3]
    next = val[loopvar] + increment
    
    if increment >= 0 then
	-- up loop
	if next > limit then
	    pc += 5 -- exit loop
	else
	    val[loopvar] = next
	    pc = Code[pc+1] -- loop again
	end if
    else    
	-- down loop
	if next < limit then
	    pc += 5 -- exit loop
	else
	    val[loopvar] = next
	    pc = Code[pc+1] -- loop again
	end if
    end if
end procedure

procedure opENDFOR_INT_UP1() 
-- ENDFOR_INT_UP1
-- faster: end of for loop with known +1 increment
-- exit or go back to the top
-- (loop var might not be integer, but that doesn't matter here)
    integer loopvar
    atom limit, next
    
    limit = val[Code[pc+2]]
    loopvar = Code[pc+3]
    next = val[loopvar] + 1
    
    -- up loop
    if next > limit then
	pc += 5 -- exit loop
    else
	val[loopvar] = next
	pc = Code[pc+1] -- loop again
    end if
end procedure

function RTLookup(sequence name, integer file, symtab_index proc, integer stlen)
-- Look up a name (routine or var) in the symbol table at runtime.
-- The name must have been defined earlier in the source than
-- where we are currently executing. The name may be a simple "name"
-- or "ns:name". Speed is not too critical. This lookup is only used 
-- in interactive trace mode, and in looking up routine id's, 
-- which should normally only be done once for an indirectly-callable 
-- routine.
    symtab_index s, global_found
    sequence ns
    integer colon
    integer ns_file

    colon = find(':', name)
    
    if colon then
	-- look up "ns : name"
	ns = name[1..colon-1]
	name = name[colon+1..$]
	
	-- trim off any trailing whitespace from ns
	while length(ns) and (ns[$] = ' ' or ns[$] = '\t') do
	    ns = ns[1..$-1]
	end while
    
	-- trim off any leading whitespace from ns
	while length(ns) and (ns[1] = ' ' or ns[1] = '\t') do
	    ns = ns[2..$]
	end while
	
	if length(ns) = 0 then
	    return 0 -- bad syntax
	end if

	-- step 1: look up NAMESPACE symbol 
	s = SymTab[TopLevelSub][S_NEXT]
	while s != 0 and s <= stlen do
	    if file = SymTab[s][S_FILE_NO] and 
		SymTab[s][S_TOKEN] = NAMESPACE and 
		equal(ns, SymTab[s][S_NAME]) then
		exit
	    end if
	    s = SymTab[s][S_NEXT]
	end while
	
	if s = 0 then
	    return 0 -- couldn't find ns
	end if
	
	ns_file = SymTab[s][S_OBJ]
	
	-- trim off any leading whitespace from name
	while length(name) and (name[1] = ' ' or name[1] = '\t') do
	    name = name[2..$]
	end while
	
	-- step 2: find global name in ns file 
	s = SymTab[TopLevelSub][S_NEXT]
	while s != 0 and s <= stlen do
	    if SymTab[s][S_FILE_NO] = ns_file and 
		SymTab[s][S_SCOPE] = SC_GLOBAL and 
		equal(name, SymTab[s][S_NAME]) then
		return s
	    end if
	    s = SymTab[s][S_NEXT]
	end while
	
	return 0 -- couldn't find name in ns file
    
    else 
	-- look up simple unqualified routine name

	if proc != TopLevelSub then  
	    -- inside a routine - check PRIVATEs and LOOP_VARs
	    s = SymTab[proc][S_NEXT]
	    while s and (SymTab[s][S_SCOPE] = SC_PRIVATE or 
			 SymTab[s][S_SCOPE] = SC_LOOP_VAR) do
		if equal(name, SymTab[s][S_NAME]) then
		    return s           
		end if
		s = SymTab[s][S_NEXT]
	    end while
	end if

	-- try to match a LOCAL or GLOBAL routine in the same source file
	s = SymTab[TopLevelSub][S_NEXT]
	while s != 0 and s <= stlen do
	    if SymTab[s][S_FILE_NO] = file and 
		(SymTab[s][S_SCOPE] = SC_LOCAL or 
		 SymTab[s][S_SCOPE] = SC_GLOBAL or 
		(proc = TopLevelSub and SymTab[s][S_SCOPE] = SC_GLOOP_VAR)) and
		equal(name, SymTab[s][S_NAME]) then  
		-- shouldn't really be able to see GLOOP_VARs unless we are
		-- currently inside the loop - only affects interactive var display
		return s
	    end if
	    s = SymTab[s][S_NEXT]
	end while 
    
	-- try to match a single earlier GLOBAL symbol
	global_found = FALSE
	s = SymTab[TopLevelSub][S_NEXT]
	while s != 0 and s <= stlen do
	    if SymTab[s][S_SCOPE] = SC_GLOBAL and 
	       equal(name, SymTab[s][S_NAME]) then
		if not global_found then
		    global_found = s
		else
		    return 0 -- 2nd global with same name
		end if
	    end if
	    s = SymTab[s][S_NEXT]
	end while 
	return global_found
    
    end if
end function

sequence e_routine -- list of routines with a routine id assigned to them
e_routine = {}

procedure opCALL_PROC() 
-- CALL_PROC, CALL_FUNC - call via routine id
    integer cf, top, n, arg
    symtab_index sub
    
    cf = Code[pc] = CALL_FUNC
    top = length(call_stack)
    
    a = Code[pc+1]  -- routine id
    if val[a] < 0 or val[a] >= length(e_routine) then
	RTFatal(165)--"invalid routine id"
    end if
    
    sub = e_routine[val[a]+1]
    b = Code[pc+2]  -- argument list
    
    if cf then
	if SymTab[sub][S_TOKEN] = PROC then
	  RTFatal(sprintf(messages[166][LANG],--"%s() does not return a value" 
			  SymTab[sub][S_NAME]))
	end if
    else
	if SymTab[sub][S_TOKEN] != PROC then
	  RTFatal(sprintf(messages[167][LANG],
	      --"the value returned by %s() must be assigned or used"
			    SymTab[sub][S_NAME]))
	end if
    end if
    if atom(val[b]) then
	RTFatal(168)--"argument list must be a sequence"
    end if
    
    if SymTab[sub][S_NUM_ARGS] != length(val[b]) then
	RTFatal(sprintf(messages[169][LANG],
	--"call to %s() via routine-id should pass %d arguments, not %d"
	{SymTab[sub][S_NAME], SymTab[sub][S_NUM_ARGS], length(val[b])}))
    end if
    
    n = SymTab[sub][S_NUM_ARGS]
    arg = SymTab[sub][S_NEXT]
    SymTab[sub][S_ACTIVE] += 1
    
    if SymTab[sub][S_ACTIVE] > 1 then
	-- recursive: push the params, privates and temps
	for i = 1 to n do
	    call_stack = append(call_stack, val[arg])
	    val[arg] = val[b][i]
	    arg = SymTab[arg][S_NEXT]
	end for
    
	-- save the privates
	while arg != 0 and SymTab[arg][S_SCOPE] <= SC_PRIVATE do
	    call_stack = append(call_stack, val[arg])
	    val[arg] = NOVALUE
	    arg = SymTab[arg][S_NEXT]
	end while
    
	-- save temps
	arg = SymTab[sub][S_TEMPS]
	while arg != 0 do
	    call_stack = append(call_stack, val[arg])
	    val[arg] = NOVALUE
	    arg = SymTab[arg][S_NEXT]
	end while
    else
	-- don't push
	-- save and copy the args
	for i = 1 to n do
	    val[arg] = val[b][i]
	    arg = SymTab[arg][S_NEXT]
	end for
    end if
    
    call_stack = append(call_stack, top)    

    pc += 3 + cf
    
    call_stack = append(call_stack, pc) 
    call_stack = append(call_stack, sub)
    
    Code = SymTab[sub][S_CODE]
    pc = 1
end procedure
	      
procedure opROUTINE_ID()
-- get the routine id for a routine name    
-- routine id's start at 0 (for compatibility with C-coded back-end)
    integer sub, fn, p, stlen
    object name
    
    sub = Code[pc+1]   -- CurrentSub
    stlen = Code[pc+2]  -- s.t. length
    name = val[Code[pc+3]]  -- routine name sequence
    fn = Code[pc+4]    -- file number
    target = Code[pc+5]
    pc += 6
    if atom(name) then
	val[target] = -1
	return
    end if
    p = RTLookup(name, fn, sub, stlen)
    if p = 0 or not find(SymTab[p][S_TOKEN], {PROC, FUNC, TYPE}) then
	val[target] = -1  -- name is not a routine
	return
    end if
    for i = 1 to length(e_routine) do
	if e_routine[i] = p then
	    val[target] = i - 1  -- routine was already assigned an id
	    return
	end if
    end for 
    e_routine = append(e_routine, p)
    val[target] = length(e_routine) - 1
end procedure
	    
procedure opAPPEND()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = append(val[a], val[b])
    pc += 4
end procedure

procedure opPREPEND() 
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = prepend(val[a], val[b])
    pc += 4
end procedure

procedure opCONCAT()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = val[a] & val[b]
    pc += 4
end procedure
	    
procedure opCONCAT_N()
-- concatenate 3 or more items
    integer n
    object x
    
    n = Code[pc+1] -- number of items
    -- operands are in reverse order
    x = val[Code[pc+2]] -- last one
    for i = pc+3 to pc+n+1 do
	x = val[Code[i]] & x
    end for
    target = Code[pc+n+2]
    val[target] = x
    pc += n+3
end procedure
	    
procedure opREPEAT()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = repeat(val[a], val[b])
    pc += 4
end procedure

procedure opDATE()
    target = Code[pc+1]
    val[target] = date()
    pc += 2
end procedure

procedure opTIME()
    target = Code[pc+1]
    val[target] = time()
    pc += 2
end procedure

procedure opSPACE_USED() -- RDS DEBUG only
    pc += 2
end procedure

procedure opNOP2()   
-- space filler
    pc+= 2
end procedure

procedure opPOSITION()
    a = Code[pc+1]
    b = Code[pc+2]
    position(val[a], val[b])  -- error checks
    pc += 3
end procedure
	    
procedure opEQUAL()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = equal(val[a], val[b])
    pc += 4
end procedure
		
procedure opCOMPARE()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = compare(val[a], val[b])
    pc += 4
end procedure

procedure opFIND()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = find(val[a], val[b])
    pc += 4
end procedure

procedure opMATCH()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = match(val[a], val[b])
    pc += 4
end procedure

procedure opPEEK4U()
    a = Code[pc+1]
    target = Code[pc+2]
    val[target] = peek4u(val[a])
    pc += 3
end procedure
	
procedure opPEEK4S()
    a = Code[pc+1]
    target = Code[pc+2]
    val[target] = peek4s(val[a])
    pc += 3
end procedure
	
procedure opPEEK()
    a = Code[pc+1]
    target = Code[pc+2]
    val[target] = peek(val[a])
    pc += 3
end procedure

procedure opPOKE()
    a = Code[pc+1]
    b = Code[pc+2]
    poke(val[a], val[b])
    pc += 3
end procedure
	
procedure opPOKE4()
    a = Code[pc+1]
    b = Code[pc+2]
    poke4(val[a], val[b])
    pc += 3
end procedure

procedure opMEM_COPY()
    a = Code[pc+1]
    b = Code[pc+2]
    c = Code[pc+3]
    mem_copy(val[a], val[b], val[c])
    pc += 4
end procedure
	    
procedure opMEM_SET()
    a = Code[pc+1]
    b = Code[pc+2]
    c = Code[pc+3]
    mem_set(val[a], val[b], val[c])
    pc += 4
end procedure
	    
procedure opPIXEL()
    a = Code[pc+1]
    b = Code[pc+2]
    pixel(val[a], val[b])
    pc += 3
end procedure
	    
procedure opGET_PIXEL()
    a = Code[pc+1]
    target = Code[pc+2]
    val[target] = get_pixel(val[a])
    pc += 3
end procedure
	  
procedure opCALL()
    a = Code[pc+1]
    call(val[a])
    pc += 2
end procedure

procedure opSYSTEM()
    a = Code[pc+1]
    b = Code[pc+2]
    system(val[a], val[b])
    pc += 3
end procedure
		
procedure opSYSTEM_EXEC()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = system_exec(val[a], val[b])
    pc += 4
end procedure
		
-- I/O routines

procedure opOPEN()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = open(val[a], val[b])
    pc += 4
end procedure

procedure opCLOSE()
    a = Code[pc+1]
    close(val[a])
    pc += 2
end procedure
	      
procedure opABORT()
    abort(val[Code[pc+1]])
end procedure

procedure opGETC()  -- read a character from a file 
    a = Code[pc+1]
    target = Code[pc+2]
    val[target] = getc(val[a])
    pc += 3
end procedure
 
procedure opGETS()  
-- read a line from a file
    a = Code[pc+1]
    target = Code[pc+2]
    val[target] = gets(val[a])
    pc += 3
end procedure

procedure opGET_KEY() 
-- read an immediate key (if any) from the keyboard 
-- or return -1 
    target = Code[pc+1]
    val[target] = get_key()
    pc += 2
end procedure

procedure opCLEAR_SCREEN()
    clear_screen()
    pc += 1
end procedure

procedure opPUTS()
    a = Code[pc+1]
    b = Code[pc+2]
    puts(val[a], val[b])
    pc += 3
end procedure

procedure opQPRINT()
-- Code[pc+1] not used
    a = Code[pc+2]
    ? val[a]
    pc += 3
end procedure
	
procedure opPRINT()
    a = Code[pc+1]
    b = Code[pc+2]
    print(val[a], val[b])
    pc += 3
end procedure

procedure opPRINTF()
    -- printf 
    a = Code[pc+1]
    b = Code[pc+2]
    c = Code[pc+3]
    printf(val[a], val[b], val[c])
    pc += 4
end procedure

procedure opSPRINTF()
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = sprintf(val[a], val[b])
    pc += 4
end procedure

procedure opCOMMAND_LINE()
    sequence cmd
    
    target = Code[pc+1]
    cmd = command_line()
    -- drop second word for better compatibility
    if length(cmd) > 2 then
	cmd = {cmd[1]} & cmd[3..$]
    end if
    val[target] = cmd
    pc += 2
end procedure

procedure opGETENV()
    a = Code[pc+1]
    target = Code[pc+2]
    val[target] = getenv(val[a])
    pc += 3
end procedure

procedure opC_PROC()
    a = Code[pc+1]
    b = Code[pc+2]
    c_proc(val[a], val[b])  -- callback could happen here
    pc += 3
end procedure
	  
procedure opC_FUNC()
    integer target
    
    a = Code[pc+1]
    b = Code[pc+2]
    target = Code[pc+3]
    val[target] = c_func(val[a], val[b])  -- callback could happen here
    pc += 4 
end procedure
	    
procedure opTRACE()
    TraceOn = val[Code[pc+1]]
    pc += 2  -- turn on/off tracing
end procedure
	    
-- other tracing/profiling ops - ignored
procedure opPROFILE() 
-- PROFILE, DISPLAY_VAR, ERASE_PRIVATE_NAMES, ERASE_SYMBOL
-- ops not implemented, ignore
    pc += 2
end procedure
	    
procedure opUPDATE_GLOBALS()
-- for interactive trace
-- not implemented, ignore
    pc += 1
end procedure

sequence operation 

procedure do_exec()
-- execute IL code, starting at pc 
    integer op

    keep_running = TRUE
    while keep_running do 
	op = Code[pc]
	call_proc(operation[op], {}) -- opcodes start at 1
    end while
    keep_running = TRUE -- so higher-level do_exec() will keep running
end procedure

--            Call-backs
--
-- This uses Intel machine code developed by Matthew Lewis.
-- It allows an "infinite" number of call-back routines to be
-- created dynamically.
--
-- Note: If you happen to port Euphoria to a non-Intel machine,
-- or a system with a different calling convention,
-- Matt's machine-code call-backs won't work, but you
-- can easily create call-back routines in Euphoria, something like:
--
--    function callback_001(atom a, atom b, atom c, atom d)
--        return general_callback(call_backs[1], {a,b,c,d})
--    end function
--
-- You can get the address of the above routine using:
--
--    addr = call_back(routine_id("call_back_001"))
--
-- By creating call_back_001, call_back_002 ... you can create
-- as many call-back routines as you like, in a portable way.
-- The only problem is that you can't dynamically create new
-- call_back routines at run-time with this method. Most programs that
-- use call-backs only need a small number of them (less than 10). 
-- 4-argument call-backs are quite common in Windows, so you might
-- need several of them on that system.


function general_callback(sequence routine, sequence args)
-- call the user's function from an external source 
-- (interface for Euphoria-coded call-backs)

    val[t_id] = routine[C_USER_ROUTINE]
    val[t_arglist] = args
    
    -- create a stack frame
    call_stack = call_stack & {length(call_stack), pc, call_back_routine} 

    Code = call_back_code 
    pc = 1 
     
    do_exec()
    
    -- remove the stack frame
    pc = call_stack[$-1]
    call_stack = call_stack[1..$-3]
    
    -- restore
    Code = SymTab[call_stack[$]][S_CODE]
    
    return val[t_return_val]
end function

forward_general_callback = routine_id("general_callback")

function machine_callback(atom cbx, atom ptr)
-- call the user's function from an external source 
-- (interface for machine-coded call-backs)
    sequence routine, args
    
    routine = call_backs[cbx]
    args = peek4u(ptr & call_backs[cbx][C_NUM_ARGS])
    
    return general_callback(routine, args)
end function

call_backs = {}

constant cb_std = {
    #89,#E0,                --    0: mov eax, esp
    #83,#C0,#04,            --    2: add eax, 4
    #50,                    --    5: push eax
    #68,#00,#00,#00,#00,    --    6: push dword rid (7)
    #FF,#15,#00,#00,#00,#00,--    B: call near dword ptr [pfunc] (13)
    #C2,#00,#00,            --   11: ret bytes (18)
    #00,#00,#00,#00},       --   14: function pointer (20)

cb_cdecl= {
    #89,#E0,                --    0: mov eax, esp
    #83,#C0,#04,            --    2: add eax, 4
    #50,                    --    5: push eax
    #68,#00,#00,#00,#00,    --    6: push dword rid (7)
    #FF,#15,#00,#00,#00,#00,--    B: call near dword ptr [pfunc] (13)
    #83, #C4, #08,          --   11: sub esp, 8
    #C3,#00,#00,            --   14: ret bytes 
	#00,#00,#00,#00}        --   17: function pointer (23)


procedure do_callback(integer b)
-- handle callback()
    symtab_index r
    atom asm
    integer id, convention
    object x
    
    -- val[b] is:  routine id or {'+', routine_id}
    x = val[b]
    if atom(x) then
	id = x
	convention = 0
    else
	id = x[2]
	convention = x[1]
    end if
	
    if id < 0 or id >= length(e_routine) then
	RTFatal(170)--"Invalid routine id"
    end if
	
    r = e_routine[id+1]

    if platform() = WIN32 and convention = 0 then
	-- stdcall
	asm = allocate( length(cb_std) )
	poke( asm, cb_std ) 
	poke4( asm + 7, length(call_backs) + 1 )
	poke4( asm + 13, asm + 20 )
	poke( asm + 18, SymTab[r][S_NUM_ARGS] * 4 )
	poke4( asm + 20, call_back( routine_id("machine_callback") ) )
	
    else
	-- cdecl
	asm = allocate( length(cb_cdecl) )  
	poke( asm, cb_cdecl )
	poke4( asm + 7, length(call_backs) + 1 )
	poke4( asm + 13, asm + 23 )
	poke4( asm + 23, call_back( '+' & routine_id("machine_callback") ) )
    end if

    val[target] = asm
    call_backs = append( call_backs, { r, id, SymTab[r][S_NUM_ARGS] })
end procedure        

procedure do_crash_routine(integer b)
-- add a crash routine to the list
    object x
    
    x = val[b]
    if atom(x) and x >= 0 and x < length(e_routine) then
	crash_list = append(crash_list, x)
    else
	RTFatal(171)--"crash routine requires a valid routine id"
    end if
end procedure

procedure opMACHINE_FUNC()
    a = Code[pc+1]
    b = Code[pc+2] 
    target = Code[pc+3]
    pc += 4
    -- handle CALL_BACK specially
    if val[a] = M_CALL_BACK then
	-- routine id's must be handled at our level
	do_callback(b)  
    else
	val[target] = machine_func(val[a], val[b])
    end if
end procedure

procedure opMACHINE_PROC()
    object v
    
    a = Code[pc+1]
    b = Code[pc+2]
    v = val[a]
    -- some things must be handled at our level, not a lower level
    if v = M_CRASH_ROUTINE then
	-- routine id's must be handled at our level
	do_crash_routine(b) 
    
    elsif v = M_CRASH_MESSAGE then
	crash_msg = val[b]
	
    elsif v = M_CRASH_FILE and sequence(val[b]) then
	err_file_name = val[b]  
    else
	machine_proc(val[a], val[b]) 
    end if
    pc += 3
end procedure
	 
global procedure InitBackEnd(integer ignore)
-- initialize Interpreter
-- Some ops are treated exactly the same as other ops.
-- In the hand-coded C back-end, they might be treated differently
-- for extra performance.
    sequence name
    
    -- set up val
    val = repeat(0, length(SymTab))
    for i = 1 to length(SymTab) do
	val[i] = SymTab[i][S_OBJ] -- might be NOVALUE
    end for
    
    -- set up operations
    operation = repeat(-1, length(opnames))
    
    for i = 1 to length(opnames) do
	name = opnames[i]
	-- some similar ops are handled by a common routine
	if find(name, {"RHS_SUBS_CHECK", "RHS_SUBS_I"}) then
	    name = "RHS_SUBS"
	elsif find(name, {"ASSIGN_SUBS_CHECK", "ASSIGN_SUBS_I"}) then
	    name = "ASSIGN_SUBS"
	elsif equal(name, "ASSIGN_I") then
	    name = "ASSIGN"
	elsif find(name, {"EXIT", "ENDWHILE"}) then
	    name = "ELSE"
	elsif equal(name, "PLUS1_I") then
	    name = "PLUS1"      
	elsif equal(name, "PRIVATE_INIT_CHECK") then
	    name = "GLOBAL_INIT_CHECK"
	elsif equal(name, "PLUS_I") then
	    name = "PLUS"
	elsif equal(name, "MINUS_I") then
	    name = "MINUS"
	elsif equal(name, "FOR_I") then
	    name = "FOR"
	elsif find(name, {"ENDFOR_UP", "ENDFOR_DOWN", 
			  "ENDFOR_INT_UP", "ENDFOR_INT_DOWN",
			  "ENDFOR_INT_DOWN1"}) then
	    name = "ENDFOR_GENERAL"
	elsif equal(name, "CALL_FUNC") then
	    name = "CALL_PROC"
	elsif find(name, {"DISPLAY_VAR", "ERASE_PRIVATE_NAMES", 
			  "ERASE_SYMBOL"}) then
	    name = "PROFILE"
	elsif equal(name, "SC2_AND") then
	    name = "SC2_OR"
	elsif find(name, {"SC2_NULL", "ASSIGN_SUBS2", "PLATFORM",
			  "END_PARAM_CHECK", "NOPWHILE", "NOP1"}) then 
	    -- never emitted
	    name = "NOP2" 
	elsif equal(name, "GREATER_IFW_I") then
	    name = "GREATER_IFW"
	elsif equal(name, "LESS_IFW_I") then
	    name = "LESS_IFW"
	elsif equal(name, "EQUALS_IFW_I") then
	    name = "EQUALS_IFW"
	elsif equal(name, "NOTEQ_IFW_I") then
	    name = "NOTEQ_IFW"
	elsif equal(name, "GREATEREQ_IFW_I") then
	    name = "GREATEREQ_IFW"
	elsif equal(name, "LESSEQ_IFW_I") then
	    name = "LESSEQ_IFW"
	end if
	
	operation[i] = routine_id("op" & name)
	if operation[i] = -1 then
	    RTInternal("no routine id for op" & name)
	end if
    end for
end procedure

global procedure Execute(symtab_index proc, integer start_index)
-- top level executor 
    CurrentSub = proc
    call_stack = {CurrentSub}
    pc = start_index
    do_exec()
end procedure

Execute_id = routine_id("Execute")

without warning
global procedure BackEnd(atom ignore)
-- The Interpreter back end
    Execute(TopLevelSub, 1)
end procedure

-- dummy routines, not used
global procedure OutputIL()
end procedure

global function extract_options(sequence s)
-- dummy routine, not used by interpreter
    return s
end function

